import sys
import json


def write_environment(water,
                      food,
                      obstacles,
                      environment_fn='environment.json'):
    """
    Write water, food, and obstacles lists to an environment file.
    """
    if len(water) != len(food) or len(water) != len(obstacles):
        sys.exit('Invalid environment arrays!')
    env_dict = {'water': water,
                'food': food,
                'obstacles': obstacles}
    with open(environment_fn, 'w') as f:
        json.dump(env_dict, f)


def constant_list(val, length):
    """
    Generate a constant-valued list.
    """
    return [val] * length



import sys
import os
import errno

import parse_intent
import dataset_io as dio
import parameter_io as pio


class Universe(object):
    """
    Create the universe of the simulation.
    """

    def __init__(self,
                 world_fn=None,
                 organisms_fn=None,
                 world_param_fn=None,
                 species_param_fns=None,
                 custom_methods_fns=None,
                 current_time=0,
                 end_time=10,
                 dataset_dir='datasets/',
                 pad_zeroes=4,
                 file_extension='.txt'):
        """
        Initialize universe based on either parameter files or saved datasets.

        Parameters
        ----------
        world_fn : str
            Filename of saved world dataset.
        organisms_fn : str
            Filename of saved organism dataset.
        world_param_fn : str
            Filename of world parameter file.
        species_param_fns : list of str
            List of filenames of species parameter files.
        custom_methods_fns : list of str
            List of filenames of external python scripts containing custom
            behaviors.
        current_time : int
            Current time of simulation.
        end_time : int
            End time of simulation.
        dataset_dir : str
            Directory path for saving all world and organism datasets.
        pad_zeroes : int
            Number of zeroes to pad in dataset filenames.
        file_extension : str
            File extension for saving dataset files. Should generally be '.txt'
            or '.json'.

        """
        self.world_fn = world_fn
        self.organisms_fn = organisms_fn
        self.world_param_fn = world_param_fn
        self.species_param_fns = species_param_fns
        self.custom_methods_fns = custom_methods_fns

        if self.custom_methods_fns is not None:
            self.custom_methods_fns = [os.path.abspath(path)
                                       for path in self.custom_methods_fns
                                       if os.path.isfile(path)]

        self.dataset_dir = dataset_dir
        if dataset_dir[-1] != '/':
            self.dataset_dir += '/'
        try:
            os.makedirs(dataset_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

        self.current_time = current_time
        self.end_time = end_time
        self.pad_zeroes = pad_zeroes
        while (self.end_time - self.current_time) >= 10 ** self.pad_zeroes:
            self.pad_zeroes += 1
        self.file_extension = file_extension

        # world is a World object
        self.world = self.initialize_world()
        # organisms is a list of Organism objects
        self.organism_list = self.initialize_organisms()
        self.intent_list = []

    def initialize_world(self):
        """
        Initialize the world of the universe from either a saved dataset
        or from a parameter file (and subsequently writing the
        initial time step to file).

        Returns
        -------
        world : World
            World at the beginning of the simulation.
        """
        if self.world_fn is not None:
            # Set up entire world based on world records
            world = dio.load_world_dataset(self.world_fn)
        elif self.world_param_fn is not None:
            # Set up entire world based on parameter file
            world = pio.load_world_parameters(self.world_param_fn)
            output_fn = (self.dataset_dir + 'world_ds'
                         + str(self.current_time).zfill(self.pad_zeroes)
                         + self.file_extension)
            dio.write_world_dataset(world, output_fn)
        else:
            sys.exit('No files specified for initialization!')
        return world

    def initialize_organisms(self):
        """
        Initialize all organisms in the universe from either a saved dataset
        or from parameter files (and subsequently writing the
        initial time step to file).

        Returns
        -------
        organism_list : list of Organisms
            List of organisms at the beginning of the simulation.
        """
        if self.organisms_fn is not None:
            # Set up all organisms based on organism records
            organism_list = dio.load_organism_dataset(self.organisms_fn)
        elif self.species_param_fns is not None:
            # Set up all organisms based on species specifications
            organism_list = pio.load_species_parameters(
                    self.species_param_fns,
                    self.world,
                    self.custom_methods_fns)
            output_fn = (self.dataset_dir + 'organisms_ds'
                         + str(self.current_time).zfill(self.pad_zeroes)
                         + self.file_extension)
            dio.write_organism_dataset(organism_list, output_fn)
        else:
            sys.exit('No files specified for initialization!')
        return organism_list

    def step(self):
        """
        Steps through one time step, iterating over all organisms and
        computing new organism states. Saves all organisms and the world
        to file at the end of each step.
        """
        self.intent_list = []
        for organism in self.organism_list:
            for new_organism in organism.step(self.organism_list, self.world):
                self.intent_list.append(new_organism)

        self.current_time += 1
        # Parse intent list and ensure it is valid
        self.organism_list = parse_intent.parse(self.intent_list,
                                                self.organism_list)

        org_output_fn = (self.dataset_dir + 'organisms_ds'
                         + str(self.current_time).zfill(self.pad_zeroes)
                         + self.file_extension)
        dio.write_organism_dataset(self.organism_list, org_output_fn)

        world_output_fn = (self.dataset_dir + 'world_ds'
                           + str(self.current_time).zfill(self.pad_zeroes)
                           + self.file_extension)
        # Potential changes to the world would go here
        dio.write_world_dataset(self.world, world_output_fn)


# At its simplest, the entire executable could just be written like this
if __name__ == '__main__':
    universe = Universe()
    while universe.current_time < universe.end_time:
        universe.step()



import uuid
import copy
import imp
import sys

import fields
from utils import cast_to_list
from organism_behavior import movement, reproduction, drinking, eating, action


class Organism(object):
    """
    A basic organism structure for all species.
    """

    def __init__(self, init_dict={}):
        """
        Create a new organism from a dictary of parameters. The dictionary
        is specified in blossom.fields.
        """
        # Sets up defaults based on organism parameters
        for (prop, default) in fields.organism_field_names.items():
            setattr(self, prop, init_dict.get(prop, default))

        # Set unique id for organism
        if self.organism_id is None:
            self.organism_id = str(uuid.uuid4())

        if self.drinking_type is not None and self.water_current is None:
            self.water_current = self.water_initial

        if self.eating_type is not None and self.food_current is None:
            self.food_current = self.food_initial

        # Import custom modules / paths
        if self.custom_methods_fns is not None:
            self.custom_modules = []
            for i, path in enumerate(cast_to_list(self.custom_methods_fns)):
                temp_module = imp.load_source('%s' % i, path)
                self.custom_modules.append(temp_module)

    @classmethod
    def clone(cls, organism):
        """
        Makes a new Organism object identical to the current one.

        Parameters
        ----------
        organism : Organism
            Organism to copy.

        Returns
        -------
        new_organism : Organism
            Copied organism.
        """
        new_organism = cls(vars(organism))
        # Use copy module to properly handle mutable lists
        new_organism.ancestry = copy.copy(new_organism.ancestry)
        new_organism.position = copy.copy(new_organism.position)
        return new_organism

    def update_parameter(self, parameter, value, method='set'):
        """
        Update a specific parameter of the organism.

        Parameters
        ----------
        parameter : string
            Parameter to update.
        value
            Value with which to update.
        method : string
            Method types are: 'set', 'add', 'subtract'.

        Returns
        -------
        self : Organism
            Same organism object with updated parameter.
        """
        attribute = getattr(self, parameter)
        if method == 'set':
            attribute = value
        elif method == 'add':
            attribute += value
        elif method == 'subtract':
            attribute -= value
        elif method == 'append':
            attribute.append(value)
        else:
            sys.exit('Invalid update method!')
        setattr(self, parameter, attribute)
        return self

    def move(self, organism_list, world):
        """
        Method for handling movement. Searches through custom methods and
        built-in movement methods.

        Parameters
        ----------
        organism_list : list of Organisms
            List of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.

        Returns
        -------
        affected_organisms : Organisms, or list of Organisms
            Organism or list of organisms affected by this organism's movement.

        """
        if self.movement_type is None:
            sys.exit('No movement type defined!')
        elif self.custom_methods_fns is not None:
            for custom_module in self.custom_modules:
                if hasattr(custom_module, self.movement_type):
                    return (getattr(custom_module, self.movement_type)
                            (self, organism_list, world))
        return (getattr(movement, self.movement_type)
                (self, organism_list, world))

    def reproduce(self, organism_list, world):
        """
        Method for handling reproduction. Searches through custom methods
        and built-in reproduction methods.

        Parameters
        ----------
        organism_list : list of Organisms
            List of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.

        Returns
        -------
        affected_organisms : Organisms, or list of Organisms
            Organism or list of organisms affected by this organism's
            reproduction. For example, this would include both parent and
            child organisms.

        """
        if self.reproduction_type is None:
            sys.exit('No reproduction type defined!')
        elif self.custom_methods_fns is not None:
            for custom_module in self.custom_modules:
                if hasattr(custom_module, self.reproduction_type):
                    return (getattr(custom_module, self.reproduction_type)
                            (self, organism_list, world))
        return (getattr(reproduction, self.reproduction_type)
                (self, organism_list, world))

    def drink(self, organism_list, world):
        """
        Method for handling drinking. Searches through custom methods and
        built-in drinking methods.

        Parameters
        ----------
        organism_list : list of Organisms
            List of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.

        Returns
        -------
        affected_organisms : Organisms, or list of Organisms
            Organism or list of organisms affected by this organism's drinking.

        """
        if self.drinking_type is None:
            sys.exit('No drinking type defined!')
        elif self.custom_methods_fns is not None:
            for custom_module in self.custom_modules:
                if hasattr(custom_module, self.drinking_type):
                    return (getattr(custom_module, self.drinking_type)
                            (self, organism_list, world))
        return (getattr(drinking, self.drinking_type)
                (self, organism_list, world))

    def eat(self, organism_list, world):
        """
        Method for handling eating. Searches through custom methods and
        built-in eating methods.

        Parameters
        ----------
        organism_list : list of Organisms
            List of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.

        Returns
        -------
        affected_organisms : Organisms, or list of Organisms
            Organism or list of organisms affected by this organism's eating.

        """
        if self.eating_type is None:
            sys.exit('No eating type defined!')
        elif self.custom_methods_fns is not None:
            for custom_module in self.custom_modules:
                if hasattr(custom_module, self.eating_type):
                    return (getattr(custom_module, self.eating_type)
                            (self, organism_list, world))
        return getattr(eating, self.eating_type)(self, organism_list, world)

    def act(self, organism_list, world):
        """
        Method that decides and calls an action for the current timestep.
        Searches through custom methods and built-in movement methods.
        The action method specifically selects an action to take, from "move",
        "reproduce", "drink", and "eat". Then the appropriate instance method
        from this class is executed to yield the final list of affect
        organisms.

        Parameters
        ----------
        organism_list : list of Organisms
            List of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.

        Returns
        -------
        affected_organisms : list of Organisms
            Organism or list of organisms affected by this organism's action.

        """
        action_name = None
        if self.custom_methods_fns is not None:
            for custom_module in self.custom_modules:
                if hasattr(custom_module, self.action_type):
                    action_name = (getattr(custom_module, self.action_type)
                                   (self, organism_list, world))
        if action_name is None:
            action_name = (getattr(action, self.action_type)
                           (self, organism_list, world))
        return cast_to_list(getattr(self, action_name)(organism_list, world))

    def update_age(self):
        """
        Increments age by 1.
        """
        return self.update_parameter('age', 1, 'add')

    def update_water(self):
        """
        Updates health parameters relevant to water consumption.

        Decreases current water level based on metabolism, and increments
        time without water accordingly. Note that organisms die of thirst if
        this reaches the maximum time without water.
        """
        self.water_current -= self.water_metabolism

        if self.water_current > self.water_capacity:
            self.water_current = self.water_capacity
        elif self.water_current <= 0:
            self.water_current = 0
            self.time_without_water += 1
        elif self.time_without_water > 0:
            self.time_without_water = 0
        return self

    def update_food(self):
        """
        Updates health parameters relevant to food consumption.

        Decreases current food level based on metabolism, and increments
        time without food accordingly. Note that organisms die of hunger if
        this reaches the maximum time without food.
        """
        self.food_current -= self.food_metabolism

        if self.food_current > self.food_capacity:
            self.food_current = self.food_capacity
        elif self.food_current <= 0:
            self.food_current = 0
            self.time_without_food += 1
        elif self.time_without_food > 0:
            self.time_without_food = 0
        return self

    def at_death(self, cause):
        """
        Check various conditions for death.

        Parameters
        ----------
        cause : str
            Potential cause of this organism's death.

        Returns
        -------
        is_dead : bool
            Returns True if organism is dead from the specified cause, False
            otherwise.
        """
        if cause == 'old_age':
            return self.age > self.max_age
        elif cause == 'thirst':
            return (self.drinking_type is not None
                    and self.time_without_water > self.max_time_without_water)
        elif cause == 'hunger':
            return (self.eating_type is not None
                    and self.time_without_food > self.max_time_without_food)
        else:
            sys.exit('Invalid cause!')

    def die(self, cause):
        """
        Method that "kills" organism.

        Parameters
        ----------
        cause : str
            Cause of this organism's death.

        Returns
        -------
        dead_organism : Organism
            New "dead" state of this organism.
        """
        self.update_parameter('alive', False)
        self.update_parameter('age_at_death', self.age)
        self.update_parameter('cause_of_death', cause)
        return self

    def step(self, organism_list, world):
        """
        Steps through one time step for this organism. Reflects changes
        based on actions / behaviors and updates to health parameters.

        Returns a list of organisms that the action produced (either new or
        altered organisms).

        Parameters
        ----------
        organism_list : list of Organisms
            List of organisms, with which this organism may interact.
        world : World
            World, with which this organism may interact.

        Returns
        -------
        affected_organisms : list of Organisms
            Organism or list of organisms affected by this organism's actions
            or health. This could be an updated version of this organism,
            especially if the organism dies during the time step, but could
            also be multiple other organisms affected by actions (i.e. children
            from reproduction).

        """
        # Create new Organism object / reference and update age
        organism = self.clone(self).update_age()
        if organism.alive:
            if not organism.at_death('old_age'):
                # Update health
                if organism.drinking_type is not None:
                    organism.update_water()
                if organism.eating_type is not None:
                    organism.update_food()

                # Keep acting if alive
                affected_organisms = organism.act(organism_list, world)

                # Check water / food status
                if organism.drinking_type is not None:
                    if organism.at_death('thirst'):
                        organism.die('thirst')
                if organism.eating_type is not None:
                    if organism.at_death('hunger'):
                        organism.die('hunger')

                return affected_organisms
            else:
                # This organism is at death from old age
                return [organism.die('old_age')]
        else:
            # Organism status already set to dead, so return organism
            # (with 'age' incremented)
            return [organism]



"""
Built-in dictionaries with world, species, and organism parameters (or fields).

Both Organism and World objects are initialized based on these field
dictionaries, and values are either populated from parameter files or take
on the default values specified in this module.
"""

world_field_names = {'dimensionality': 1,
                     'world_size': [10],
                     'initial_environment_dict': None,
                     'water': None,
                     'food': None,
                     'obstacles': None}

specific_organism_field_names = {'organism_id': None,
                                 'dna': '0000',
                                 'age': 0,
                                 'alive': True,
                                 'age_at_death': None,
                                 'cause_of_death': None,
                                 'position': [0],
                                 'sex': None,
                                 'ancestry': [],
                                 'custom_methods_fns': None,
                                 'water_current': None,
                                 'time_without_water': 0,
                                 'food_current': None,
                                 'time_without_food': 0}

species_field_names = {'species_name': 'species1',
                       'movement_type': 'stationary',
                       'reproduction_type': None,
                       'drinking_type': None,
                       'eating_type': None,
                       'action_type': 'move_only',
                       'dna_length': 4,
                       'max_age': 20,
                       'max_time_without_food': None,
                       'max_time_without_water': None,
                       'mutation_rate': None,
                       'food_capacity': None,
                       'food_initial': None,
                       'food_metabolism': None,
                       'food_intake': None,
                       'water_capacity': None,
                       'water_initial': None,
                       'water_metabolism': None,
                       'water_intake': None}

# Combine organism and species field names
organism_field_names = dict(specific_organism_field_names,
                            **species_field_names)



import fields


class World(object):
    """
    World class for the environment of the simulation.
    """

    def __init__(self, init_dict={}):
        """
        Create a new world from a dictary of parameters. The dictionary
        is specified in blossom.fields.
        """
        for (prop, default) in fields.world_field_names.items():
            setattr(self, prop, init_dict.get(prop, default))



"""
Load information from parameter files and construct world and
organism objects at the initial timestep.
"""

import glob
import configparser
import json
import random
import copy

from utils import cast_to_list
from world import World
from organism import Organism


def load_world_parameters(fn):
    """
    Load world parameter file and construct initial World object.

    Parameters
    ----------
    fn : str
        Input filename of parameter file.

    Returns
    -------
    world : World
        World object constructed from the parameter file.
    """

    env_file = glob.glob(fn)
    if len(env_file) == 0:
        raise IndexError('There is no environment configuration file in the '
                         + 'current directory.')
    if len(env_file) > 1:
        print('There are multiple environment configuration files in the '
              + 'current directory. There should only be one environment '
              + 'configuration file. Selecting the first provided..')
    env_file = env_file[0]

    world_dict = {}
    # Load from config file
    config_world = configparser.ConfigParser()
    config_world.read(env_file)

    # environment_filename: str, or 'None'
    environment_filename = config_world.get('Overall Parameters',
                                            'environment_filename')
    if environment_filename != 'None':
        with open(environment_filename, 'r') as f:
            initial_environment_dict = json.load(f)

        world_dict['water'] = initial_environment_dict['water']
        world_dict['food'] = initial_environment_dict['food']
        world_dict['obstacles'] = initial_environment_dict['obstacles']

        if type(world_dict['water'][0]) is list:
            world_dict['dimensionality'] = 2
            world_dict['world_size'] = [len(world_dict['water']),
                                        len(world_dict['water'][0])]
        else:
            world_dict['dimensionality'] = 1
            world_dict['world_size'] = [len(world_dict['water'])]

    else:
        # world_size: space delimited ints, or 'None'
        # example: world_size = 10 10
        world_size = config_world.get('Overall Parameters', 'world_size')
        if world_size == 'None':
            # this is probably harder to do,
            # since we'd have to allow for infinite bounds...
            world_size = None
            return -1
        else:
            world_size = [int(L) for L in world_size.split()]
            world_dict['world_size'] = world_size
            world_dict['dimensionality'] = len(world_size)

            if len(world_size) == 2:
                blank_vals = [[0 for x in range(world_size[1])]
                              for x in range(world_size[0])]
            elif len(world_size) == 1:
                blank_vals = [0 for x in range(world_size)[0]]
            else:
                raise ValueError

            world_dict['water'] = copy.deepcopy(blank_vals)
            world_dict['food'] = copy.deepcopy(blank_vals)
            world_dict['obstacles'] = copy.deepcopy(blank_vals)

    return World(world_dict)


def load_species_parameters(fns, init_world, custom_methods_fns):
    """
    Load all available species parameter files.

    Parameters
    ----------
    fns : list of str
        Input filenames of species parameter files. Different species get
        different species parameter files, from which the individual organisms
        are initialized.
    init_world : World
        Initial World instance for this Universe.
    custom_methods_fns : list of str
        List of external Python scripts containing custom organism
        behaviors. :mod:`blossom` will search for methods within each
        filename included here.

    Returns
    -------
    organism_list : list of Organisms
        A list of Organism objects constructed from the parameter file.

    """

    # Find organism filenames, can be a list of patterns
    fns = cast_to_list(fns)
    org_files = [fn for pattern in fns for fn in glob.glob(pattern)]

    # Initialize list of dictionaries to hold all organism parameters
    # Each dictionary contains parameters for a single species
    organism_list = []
    for org_file in org_files:
        # Initialize temporary dict to store species parameters
        organism_dict = {}

        # Parameters that must be str
        param_str = ['species_name',
                     'movement_type',
                     'reproduction_type',
                     'drinking_type',
                     'action_type',
                     'eating_type']

        # Parameters that must be int
        param_int = ['population_size',
                     'dna_length',
                     'max_age']

        # Parameters that must be int or 'None'
        param_int_none = ['max_time_without_food',
                          'max_time_without_water',
                          'mutation_rate',
                          'food_capacity',
                          'food_initial',
                          'food_metabolism',
                          'food_intake',
                          'water_capacity',
                          'water_initial',
                          'water_metabolism',
                          'water_intake']

        # Parameters that must be float
        param_float = ['proportion_m',
                       'proportion_f',
                       'proportion_a',
                       'proportion_h']

        # Parameters that must be bool
        param_bool = ['can_mutate']

        # Load from config file
        config_org = configparser.ConfigParser()
        config_org.read(org_file)

        # Cycle through all parameters in the config file,
        # converting them to proper types as specifed above
        for section in config_org.sections():
            for (key, val) in config_org.items(section):

                if key in param_str:
                    pass
                elif key in param_int:
                    val = int(val)
                elif key in param_int_none:
                    if val != 'None':
                        val = int(val)
                elif key in param_float:
                    val = float(val)
                elif key in param_bool:
                    val = (val == 'True')
                else:
                    print(key, val)
                    raise NameError('Key is not a valid parameter')
                # ensure 'None' parameters are set to None
                if val == 'None':
                    val = None
                organism_dict[key] = val

        # Track custom method file paths
        organism_dict['custom_methods_fns'] = custom_methods_fns

        # Generate all organisms
        for i in range(organism_dict['population_size']):
            # Vary organism location randomly
            position = []
            for i in range(init_world.dimensionality):
                position.append(random.randrange(0, init_world.world_size[i]))
            organism_dict['position'] = position

            # Add organism to organism list
            organism_list.append(Organism(organism_dict))

        return organism_list



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='blossom',
    version='1.1.0',
    author='Bryan Brzycki',
    author_email='bbrzycki@berkeley.edu',
    description='A simple evolution simulator',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/blossom-evolution/blossom',
    project_urls={
        'Documentation': 'https://blossom.readthedocs.io/en/latest/',
        'Source': 'https://github.com/blossom-evolution/blossom'
    },
    packages=setuptools.find_packages(),
    install_requires=[
       'uuid',
       'numpy',
       'configparser'
    ],
    classifiers=(
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ),
)



"""
Load information from a certain dataset, e.g. to resume a simulation, and
write world and organism data back to file.
"""

import json

from world import World
from organism import Organism

def load_world_dataset(fn):
    """
    Load dataset file from JSON.
    filenames can be a single string or a list of strings.

    Parameters
    ----------
    fn : str
        Input filename of saved world dataset.

    Returns
    -------
    world : World
        World object reconstructed from the saved dataset.
    """
    with open(fn, 'r') as f:
        world_dict = json.load(f)
    return World(world_dict)

def write_world_dataset(world, fn):
    """
    Write world information from World object to file in JSON format.

    Parameters
    ----------
    world : World
        World attributes to write to file.
    fn : str
        Output filename of saved world dataset.
    """
    with open(fn, 'w') as f:
        json.dump(vars(world), f, indent=2)

def load_organism_dataset(fn):
    """
    Load dataset file from JSON.
    filenames can be a single string or a list of strings.

    Parameters
    ----------
    fn : str
        Input filename of saved organism dataset.

    Returns
    -------
    organism_list : list of Organisms
        A list of Organism objects reconstructed from the saved dataset.
    """
    with open(fn, 'r') as f:
        organism_dict_list = json.load(f)
    organism_list = [Organism(organism_dict) for organism_dict in organism_dict_list]
    return organism_list

def write_organism_dataset(organism_list, fn):
    """
    Write organism data from list of Organism objects to file in JSON
    format.

    Parameters
    ----------
    organism_list : list of Organisms
        List of Organisms to write to file.
    fn : str
        Output filename of saved organism dataset.
    """
    organism_dict_list = []
    for organism in organism_list:
        organism_dict = vars(organism)
        # Make sure we're not serializing the loaded modules themselves
        del organism_dict['custom_modules']
        organism_dict_list.append(organism_dict)
    with open(fn, 'w') as f:
        json.dump(organism_dict_list, f, indent=2)



import uuid
import math


def pure_replication(organism, organism_list, world):
    """
    Replace organism with two organism with similar parameters.
    Essentially, only differences in parameters are organism id,
    ancestry, age, and water / food levels.
    """

    new_organism_list = []

    # Generate new organisms
    for i in range(2):
        child = organism.clone(organism)
        child.update_parameter('age', 0)
        child.update_parameter('organism_id', str(uuid.uuid4()))
        child.update_parameter('ancestry', organism.organism_id, 'append')
        if organism.drinking_type is not None:
            child.update_parameter('water_current',
                                   math.floor(organism.water_current / 2))
        if organism.eating_type is not None:
            child.update_parameter('food_current',
                                   math.floor(organism.food_current / 2))
        new_organism_list.append(child)
    return new_organism_list



"""
Common utilities used throughout :mod:`blossom`
"""

def cast_to_list(x):
    """
    Make a list out of the input if the input isn't a list.
    """
    if type(x) is list:
        return x
    else:
        return [x]



import sys


def constant_eat(organism, organism_list, world):
    """
    Intake constant amount of food from world if food is present.
    """
    size = world.world_size

    if len(size) == 1:
        available_food = world.food[organism.position[0]]
    elif len(size) == 2:
        available_food = (world.food[organism.position[0]]
                                    [organism.position[1]])
    else:
        sys.exit('Invalid world dimensionality: %s' % len(size))

    diff = organism.food_capacity - organism.food_current
    intake = min(available_food, diff, organism.food_intake)

    organism.food_current += intake
    if len(size) == 1:
        world.food[organism.position[0]] -= intake
    else:
        world.food[organism.position[0]][organism.position[1]] -= intake

    return [organism]
